using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VVVV.PluginInterfaces.V2;
using System.IO;
using VVVV.Utils.Streams;
using VVVV.Utils.Linq;

namespace VVVV.Nodes.Raw
{
    [PluginInfo(Name = "+", Category = "Raw", Version = "Spectral")]
    public class PlusSpectralNode : IPluginEvaluate
    {
        [Input("Input")]
        public ISpread<ISpread<Stream>> FInputs;

        [Input("Intersperse Sequence", Order = int.MaxValue, IsSingle = true)]
        public ISpread<Stream> FIntersperseSequence;

    	[Input("Every nth", IsSingle = true, DefaultValue = 1, MinValue = 1)]
        public ISpread<int> FDownsampling;
    	
    	[Input("Bytes per Element", IsSingle = true, DefaultValue = 4, MinValue = 1)]
        public ISpread<int> FSizePerElement;
    	
        [Output("Output")]
        public ISpread<Stream> FOutput;

        public void Evaluate(int spreadMax)
        {
            FOutput.SliceCount = FInputs.SliceCount;
            var intersperseElement = FIntersperseSequence[0];
            for (int i = 0; i < FOutput.SliceCount; i++)
            {
                IEnumerable<Stream> streams = FInputs[i];
                if (intersperseElement.Length > 0)
                    streams = streams.Intersperse(intersperseElement);
                FOutput[i] = new AggregatedStreamAdvanced(streams,FSizePerElement[0], FDownsampling[0]);
            }
        }
    }
	
	public class AggregatedStreamAdvanced : ComStream
    {
        private readonly List<Tuple<long, Stream>> cache = new List<Tuple<long, Stream>>();
        private long length;
        private int BytesPerElement;
        private int EveryNth;

        public AggregatedStreamAdvanced(IEnumerable<Stream> streams, int bytesPerElement, int everyNth)
        {
            foreach (var s in streams)
            {
                var t = Tuple.Create(s.Length, s);
                this.length += s.Length;
                this.cache.Add(t);
            }
            this.length = (long)Math.Floor((double)this.Length / (double)everyNth);
            BytesPerElement = bytesPerElement;
            EveryNth = everyNth;
        }

        public override bool CanRead
        {
            get { return true; }
        }

        public override bool CanSeek
        {
            get { return true; }
        }

        public override bool CanWrite
        {
            get { return false; }
        }

        public override void Flush()
        {
            throw new NotSupportedException();
        }

        public override long Length
        {
            get { return this.length; }
        }

        public override long Position
        {
            get;
            set;
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            long streamOffset = 0;
            var totalBytesRead = 0;
            foreach (var tuple in this.cache)
            {
                var streamLength = tuple.Item1;
                var stream = tuple.Item2;

                if ((streamOffset <= (Position * EveryNth)) && ((Position * EveryNth) < streamOffset + streamLength))
                {
                    stream.Position = (Position * EveryNth) - streamOffset;
                    var bytesToRead = count - totalBytesRead;
                    
                    while (bytesToRead > 0)
                    {
                        if (Position >= length) break;

                        if( Math.Floor( (double) stream.Position / (double) BytesPerElement) % EveryNth == 0)
                        {
                            var bytesRead = stream.Read(buffer, offset + totalBytesRead, 1);
                            if (bytesRead == 0) break;
                            Position += bytesRead;
                            bytesToRead -= bytesRead;
                            totalBytesRead += bytesRead;
                        }
                        else
                        {
                            stream.Position+= 1;
                        }
                        
                    }
                }
                streamOffset += streamLength;
            }
            return totalBytesRead;
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            switch (origin)
            {
                case SeekOrigin.Begin:
                    Position = offset;
                    break;
                case SeekOrigin.Current:
                    Position += offset;
                    break;
                case SeekOrigin.End:
                    Position = Length - offset;
                    break;
                default:
                    throw new NotImplementedException();
            }
            return Position;
        }

        public override void SetLength(long value)
        {
            throw new NotSupportedException();
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            throw new NotSupportedException();
        }
    }
}